# -*- coding: utf-8 -*-

__author__ = r'wsb310@gmail.com'

import os
import sys
import asyncio

from ... import hagworm_slogan
from ... import __version__ as hagworm_version

from .base import install_uvloop, Utils

from ..error import catch_error
from ..process import fork_processes
from ..logging import DEFAULT_LOG_FILE_ROTATOR, DEFAULT_LOG_FILE_NAME, init_logger
from ..interface import FunctorInterface, RunnableInterface
from ..asyncio.zmq import Push, Pull


SIGNAL_PROTOCOL = r'tcp'
SIGNAL_PORT_1 = 83310
SIGNAL_PORT_2 = 10601


class Daemon(FunctorInterface):

    def __init__(self, child_num=2, keep_alive=False):

        if child_num < 2:
            raise ValueError(r'child_num cannot be less than 2')

        self._child_num = child_num
        self._keep_alive = keep_alive

        self._pids = set()

    def is_active(self):

        return self._child_num != 0 and len(self._pids) != 0

    def _fork_processes(self):

        while len(self._pids) < self._child_num:

            pid = fork_processes()

            if pid == 0:
                self._child_num = 0
                return False

            self._pids.add(pid)

        return True

    def _check_processes(self):

        for pid in self._pids.copy():

            if os.waitpid(pid, os.WNOHANG)[0] != pid:
                continue

            self._pids.remove(pid)

            if self._keep_alive:
                self._fork_processes()

    def _run(self):

        while True:

            self._check_processes()

            if not self.is_active():
                return

    def __call__(self):

        if self._fork_processes():
            self._run()
            sys.exit(0)


class MainProcessAbstract(Daemon):

    def __init__(self, child_num=2):

        Daemon.__init__(self, child_num, False)

        global SIGNAL_PROTOCOL, SIGNAL_PORT_1, SIGNAL_PORT_2

        self._push_server = Push(f'{SIGNAL_PROTOCOL}://*:{SIGNAL_PORT_1}', True)
        self._pull_server = Pull(f'{SIGNAL_PROTOCOL}://*:{SIGNAL_PORT_2}', True)

    async def _handle_message(self, message):
        raise NotImplementedError()

    async def _recv_message(self):

        while True:

            message = await self._pull_server.recv(True)

            if message is None:
                break
            else:
                await self._handle_message(message)

    async def _run(self):

        while True:

            self._check_processes()

            if self.is_active():
                await self._recv_message()
            else:
                break

    def __call__(self):

        if not self._fork_processes():
            return

        self._push_server.open()
        self._pull_server.open()

        with catch_error():
            Utils.run_until_complete(self._run())

        self._push_server.close()
        self._pull_server.close()

        sys.exit(0)


class ChildProcessAbstract(FunctorInterface):

    def __init__(self):

        global SIGNAL_PROTOCOL, SIGNAL_PORT_1, SIGNAL_PORT_2

        self._push_client = Push(f'{SIGNAL_PROTOCOL}://localhost:{SIGNAL_PORT_2}')
        self._pull_client = Pull(f'{SIGNAL_PROTOCOL}://localhost:{SIGNAL_PORT_1}')

        self._process_id = Utils.getpid()

    async def __call__(self):

        self._push_client.open()
        self._pull_client.open()

        with catch_error():
            await self._run()

        self._push_client.close()
        self._pull_client.close()

    async def _run(self):
        raise NotImplementedError()


class Launcher(RunnableInterface):
    """异步版本的启动器

    用于简化和统一程序的启动操作

    """

    def __init__(self,
                 log_level=r'info', log_handler=None, log_file_path=None, log_file_name=DEFAULT_LOG_FILE_NAME,
                 log_file_rotation=DEFAULT_LOG_FILE_ROTATOR, log_file_retention=0xff, log_extra=None, log_enqueue=False,
                 daemon=None, debug=False
                 ):

        init_logger(
            level=log_level.upper(),
            handler=log_handler,
            file_path=log_file_path,
            file_name=log_file_name,
            file_rotation=log_file_rotation,
            file_retention=log_file_retention,
            extra=log_extra,
            enqueue=log_enqueue,
            debug=debug
        )

        environment = Utils.environment()

        Utils.log.info(
            f'{hagworm_slogan}'
            f'hagworm {hagworm_version}\n'
            f'python {environment["python"]}\n'
            f'system {" ".join(environment["system"])}'
        )

        install_uvloop()

        if daemon:
            daemon()

        self._event_loop = asyncio.get_event_loop()
        self._event_loop.set_debug(debug)

    def run(self, func, *args, **kwargs):

        pid = Utils.getpid()

        Utils.log.success(f'Start process pid.{pid}')

        self._event_loop.run_until_complete(func(*args, **kwargs))

        Utils.log.success(f'Stop process pid.{pid}')



# -*- coding: utf-8 -*-

__author__ = r'wsb310@gmail.com'

import zmq

from zmq.asyncio import Context

from ..interface import ContextManager

from .base import Utils, AsyncCirculator
from .buffer import QueueBuffer


class _SocketBase(ContextManager):

    def __init__(self, name, socket_type, address, bind_mode):

        self._name = name if name else Utils.uuid1()[:8]

        self._context = Context.instance()

        self._socket = None
        self._socket_type = socket_type

        self._address = address
        self._bind_mode = bind_mode

    def _context_initialize(self):

        self.open()

    def _context_release(self):

        self.close()

    @property
    def socket(self):

        return self._socket

    def open(self, hwm=None):

        self._socket = self._context.socket(self._socket_type)

        if self._bind_mode:
            self._socket.bind(self._address)
        else:
            self._socket.connect(self._address)

        if hwm is not None:
            self._socket.set_hwm(hwm)

    def close(self):

        if self._socket and not self._socket.closed:
            self._socket.close()
            self._socket = None

    async def send(self, data, noblock=False):

        result = False

        try:

            if noblock is True:
                await self._socket.send_pyobj(data, zmq.NOBLOCK)
            else:
                await self._socket.send_pyobj(data)

        except zmq.error.Again as _:
            pass

        else:
            result = True

        return result

    async def recv(self, noblock=False):

        result = None

        try:

            if noblock is True:
                result = await self._socket.recv_pyobj(zmq.NOBLOCK)
            else:
                result = await self._socket.recv_pyobj()

        except zmq.error.Again as _:
            pass

        return result


##################################################
# 订阅模式

class Subscribe(_SocketBase):

    def __init__(self, address, bind_mode=False, *, name=None, topic=r''):

        super().__init__(name, zmq.SUB, address, bind_mode)

        self._msg_topic = topic

    def open(self, hwm=None):

        super().open(hwm)

        self._socket.setsockopt_string(zmq.SUBSCRIBE, self._msg_topic)


class Publish(_SocketBase):

    def __init__(self, address, bind_mode=False, *, name=None):

        super().__init__(name, zmq.PUB, address, bind_mode)


class PublishWithBuffer(_SocketBase):

    def __init__(self, address, bind_mode=False, *, name=None, buffer_slice_size=0xffff, buffer_timeout=1):

        _SocketBase.__init__(self, name, zmq.PUB, address, bind_mode)

        self._buffer = QueueBuffer(self._handle_data, buffer_slice_size, buffer_timeout)

    async def _handle_data(self, data):

        await self._socket.send_pyobj(data)

    def send(self, data):

        self._buffer.append(data)

    async def safe_close(self, timeout=0):

        async for _ in AsyncCirculator(timeout):
            if self._buffer.size() == 0:
                super().close()
                break


##################################################
# 请求模式

class Request(_SocketBase):

    def __init__(self, address, *, name=None):

        super().__init__(name, zmq.REQ, address, False)


class Reply(_SocketBase):

    def __init__(self, address, *, name=None):

        super().__init__(name, zmq.REP, address, True)


##################################################
# 推拉模式

class Push(_SocketBase):

    def __init__(self, address, bind_mode=False, *, name=None):

        super().__init__(name, zmq.PUSH, address, bind_mode)


class Pull(_SocketBase):

    def __init__(self, address, bind_mode=False, *, name=None):

        super().__init__(name, zmq.PULL, address, bind_mode)



# -*- coding: utf-8 -*-

__author__ = r'wsb310@gmail.com'

import aio_pika

from ...extend.asyncio.task import IntervalTask


class RabbitMQPool:
    """RabbitMQPool连接池
    """

    def __init__(self, url, pool_size, **kwargs):

        self._mq_url = url
        self._mq_setting = kwargs

        self._pool_size = pool_size

        self._connection_pool = aio_pika.pool.Pool(self._create_connection, max_size=pool_size)
        self._channel_pool = aio_pika.pool.Pool(self._create_channel, max_size=pool_size)

    async def _create_connection(self):

        return await aio_pika.connect_robust(self._mq_url, **self._mq_setting)

    async def _create_channel(self):

        async with self._connection_pool.acquire() as connection:
            return await connection.channel()

    async def close(self):

        await self._channel_pool.close()
        await self._connection_pool.close()

    def acquire_connection(self):

        return self._connection_pool.acquire()

    def acquire_channel(self):

        return self._channel_pool.acquire()


class RabbitMQProducer(RabbitMQPool):
    """RabbitMQ发布者
    """

    async def publish(self, message, routing_key, **kwargs):

        async with self.acquire_channel() as channel:
            await channel.default_exchange.publish(aio_pika.Message(message), routing_key, **kwargs)

    async def batch_publish(self, messages, routing_key, **kwargs):

        async with self.acquire_channel() as channel:
            for message in messages:
                await channel.default_exchange.publish(aio_pika.Message(message), routing_key, **kwargs)


class RabbitMQConsumer(RabbitMQPool):
    """RabbitMQ消费者
    """

    def __init__(self, url, pool_size, **kwargs):

        super().__init__(url, pool_size, **kwargs)

        self._queue_name = None
        self._consume_func = None
        self._consume_qos = None

        self._task = IntervalTask.create(10, self._consume_message)

    async def open(self, queue_name, consume_func, consume_qos=10):

        self._queue_name = queue_name
        self._consume_func = consume_func
        self._consume_qos = consume_qos

        self._task.start()

    async def close(self):

        self._task.stop()

        await super().close()

    async def _consume_message(self):

        async with self.acquire_channel() as channel:

            await channel.set_qos(self._consume_qos)

            queue = await channel.declare_queue(self._queue_name)

            async with queue.iterator() as queue_iter:
                async for message in queue_iter:
                    await self._consume_func(message)
                    await message.ack()



# -*- coding: utf-8 -*-

__author__ = r'wsb310@gmail.com'

import ssl
import setuptools

from hagworm import __version__


ssl._create_default_https_context = ssl._create_unverified_context

with open(r'README.md', encoding=r'utf8') as stream:
    long_description = stream.read()

setuptools.setup(
    name=r'hagworm',
    version=__version__,
    license=r'Apache License Version 2.0',
    platforms=[r'all'],
    author=r'Shaobo.Wang',
    author_email=r'wsb310@gmail.com',
    description=r'Network Development Suite',
    long_description=long_description,
    long_description_content_type=r'text/markdown',
    url=r'https://gitee.com/wsb310/hagworm',
    packages=setuptools.find_packages(),
    package_data={r'hagworm': [r'static/*.*']},
    python_requires=r'>= 3.8',
    install_requires=[
        r'APScheduler==3.9.1',
        r'Pillow==9.0.1',
        r'PyJWT==2.3.0',
        r'PyYAML==6.0',
        r'SQLAlchemy==1.3.24',
        r'aiohttp==3.8.1',
        r'aiomysql==0.1.1',
        r'aioredis==1.3.1',
        r'aioredis-cluster==1.7.1',
        r'aiosmtplib==1.1.6',
        r'aio-pika==8.0.3',
        r'aliyun-log-python-sdk==0.7.9',
        r'cachetools==5.0.0',
        r'cryptography==37.0.2',
        r'etcd3gw==1.0.2',
        r'fastapi==0.76.0',
        r'fastapi-health==0.4.0',
        r'gunicorn==20.1.0',
        r'hiredis==2.0.0',
        r'httptools==0.4.0',
        r'loguru==0.6.0',
        r'motor==3.0.0',
        r'mq-http-sdk==1.0.3',
        r'msgpack==1.0.3',
        r'ntplib==0.4.0',
        r'numpy==1.22.3',
        r'psutil==5.9.0',
        r'pyahocorasick==1.4.4',
        r'pytest-asyncio==0.18.3',
        r'python-stdnum==1.17',
        r'python-multipart==0.0.5',
        r'pyzmq==22.3.0',
        r'qrcode==7.3.1',
        r'terminal-table==2.0.1',
        r'ujson==5.2.0',
        r'uvicorn[standard]==0.17.6',
        r'uvloop==0.16.0;sys_platform!="win32"',
        r'xmltodict==0.13.0',
    ],
    classifiers=[
        r'Programming Language :: Python :: 3.8',
        r'License :: OSI Approved :: Apache Software License',
        r'Operating System :: POSIX :: Linux',
    ],
    ext_modules=[
        setuptools.Extension(
            r'hagworm.extend.math', [r'./c_extend/math.c']
        ),
    ],
)



# -*- coding: utf-8 -*-

__author__ = r'wsb310@gmail.com'

from etcd3gw.client import Etcd3Client

from ...extend.asyncio.pool import ObjectPool


class Etcd3ClientPool(ObjectPool):

    def __init__(self, host, port=2379, *, maxsize=100, api_path=r'/v3/', **kwargs):

        self._setting = dict(host=host, port=port, api_path=api_path, **kwargs)

        super().__init__(maxsize)

    def _create_obj(self):

        return Etcd3Client(**self._setting)


